from enum import Enum
from typing import List

class QuestionProperties:

    class QuestionType(Enum):

        MULTIPLE_CHOICE = 'M', 'Multiple choice'
        LIST = 'L', 'List'
        RANKING = 'R', 'Ranking'
        LONG_FREE_TEXT = 'T', 'Long free text'
        NUMERICAL_INPUT = 'N', 'Numerical input'
        MATRIX = 'F', 'Matrix'
        SHORT_FREE_TEXT = 'S', 'Short free text'

    @staticmethod
    def get_question_type_from_string(type: str):
        for qt in QuestionProperties.QuestionType:
            if qt.value[0] == type:
                return qt
        return None


from typing import Union, Tuple, List, Any
import pandas as pd

from .question import Question

class ResponseSet:

    def __init__(self, question: Question):
        self._question = question
        self._items = None
        self._other_items = None
        self._response_objects = None

    @property
    def items(self) -> pd.DataFrame:
        return self._items
    @items.setter
    def items(self, value: pd.DataFrame):
        self._items = value 

    @property
    def other_items(self) -> pd.DataFrame:
        return self._other_items
    @other_items.setter
    def other_items(self, value: pd.DataFrame):
        self._other_items = value

    @property
    def question(self) -> Question:
        return self._question
    @question.setter
    def question(self, value: Question):
        self._question = value

    @property
    def response_objects(self) -> Any:
        return self._response_objects
    @response_objects.setter
    def response_objects(self, value: Any):
        self._response_objects = value


from .survey_item import SurveyItem

class ItemContainer:

    def __init__(self):
        self.items = []

    def __repr__(self):
        str = "ItemSet:\n"
        for item in self.items:            
           str += f"SurveyItem (code: {item.code}, id: {item.id}, parent_id: {item.parent_id})\n"
        return str

    def get_all(self) -> list[SurveyItem]:
        return self.items

    def add_item(self, item: SurveyItem):
        self.items.append(item)

    def get_item_by_code(self, code: str) -> SurveyItem:
        for q in self.items:
            if q.code == code:
                return q
        return None
    
    def get_item_by_id(self, qid: str) -> SurveyItem:
        for q in self.items:
            if q.qid == qid:
                return q
        return None
    



from .item_set import ItemContainer
from .question import Question

from typing import List

class SurveyStructure:

    def __init__(self, structure_filename: str):
        self.structure_filename = structure_filename
        self.question_set = ItemContainer()
        self.survey_title = {}


    def add_question(self, question: Question):
        self.question_set.add_item(question)

    def add_title(self, lang: str, title: str):
        self.survey_title[lang] = title

    def get_title(self, lang: str) -> str:
        if lang in self.survey_title:
            return self.survey_title[lang]
        return None
    
    def get_titles(self) -> dict:
        return self.survey_title

    def get_question_by_id(self, qid: str) -> Question:
        q: Question
        for q in self.question_set.get_all():
            if q.id == qid:
                return q
        return None
    
    def get_question_by_code(self, code: str) -> Question:
        q : Question
        for q in self.question_set.get_all():
            if q.code == code:
                return q
        return None

    def get_questions(self) -> List[Question]:
        return self.question_set.get_all()
    




import altair as alt
import pandas as pd
from typing import List
from matplotlib.colors import LinearSegmentedColormap, to_hex, Normalize
from matplotlib import cm
import matplotlib.pyplot as plt
import numpy as np
from enum import Enum
import deep_translator as dptr
import seaborn as sns

from .question import Question


class ChartType(Enum):
    BAR = 'multiple_choice'
    STACKED_MATRIX = 'matrix'
    RANKS = 'ranked_items'
    FREQUENCY = 'frequency'
    HISTOGRAM = 'histogram'
    PIE = 'piechart'
    BOXPLOT = 'boxplot'

 
class Charting:


    # try to set default renderer
    renderer: str = 'html'
        
    """Default Altair display frontend, respectively, rendering API to use for displaying plots."""

    translator = dptr.GoogleTranslator(source='auto', target='en')
    """Translation for chart labels."""

    @classmethod
    def set_display_frontend(cls, renderer: str):
        """Set the Altair renderer to use for displaying charts.

        :param renderer: Renderer to use, see https://altair-viz.github.io/user_guide/display_frontends.html
        :type renderer: str
        """
        alt.renderers.enable(renderer)

    @staticmethod
    def plot_chart(type: ChartType, question: Question, items, title: str, lang: str = "en", plot_labels: bool = None, **kwargs):
                        
        if type.value == ChartType.BAR.value:
            Charting.plot_bars(items, title=title, lang=lang, plot_labels=plot_labels, **kwargs)
        elif type.value == ChartType.STACKED_MATRIX.value:
            Charting.plot_stacked_matrix(items, title=title, lang=lang, plot_labels=plot_labels, **kwargs)
        elif type.value == ChartType.RANKS.value:
            Charting.plot_rank(items, title=title, lang=lang, plot_labels=plot_labels, **kwargs)
        elif type.value == ChartType.FREQUENCY.value:
            Charting.plot_frequency(items, title=title, lang=lang, **kwargs)
        elif type.value == ChartType.HISTOGRAM.value:
            Charting.plot_histogram(items, title=title, lang=lang, **kwargs)
        elif type.value == ChartType.PIE.value:
            Charting.plot_pie(items, title=title, lang=lang, plot_labels=plot_labels, **kwargs)
        elif type.value == ChartType.BOXPLOT.value:
            Charting.plot_boxplot(items, title=title, lang=lang, **kwargs)
        else:
            print(f"Chart type {type} not recognized.")

    @staticmethod
    def plot_boxplot(items: pd.DataFrame, title: str, lang: str, **kwargs):

        rank_label = 'Ranked items count'
        Charting.translator.target = lang
        if lang != 'en':
            rank_label = Charting.translator.translate(rank_label)

        box = alt.Chart(data=items).encode(
            alt.X('count:Q', title=rank_label),
        ).mark_boxplot(
            color='#222222'
        )

        items['mean_val'] = items['count'].mean()
        
        mean_val = alt.Chart(data=items).encode(
            x='mean_val:Q'
        ).mark_circle(
            color='white'
        )

        (box + mean_val).properties(
            title = alt.Title(
                title, 
                anchor="start", 
                orient="bottom", 
                offset=10)
        ).show()

    @staticmethod
    def plot_frequency(items: pd.DataFrame, title: str, lang: str = "en", maxbins: int = 10, **kwargs):

        # per-chart translation
        Charting.translator.target = lang
        x_axis_title = 'Response' if lang == 'en' else Charting.translator.translate('Response')
        y_axis_title = 'Total count' if lang == 'en' else Charting.translator.translate('Total count')
        
        hist = alt.Chart(items).mark_bar().encode(
            alt.X('value', bin=alt.Bin(maxbins=maxbins), title=x_axis_title),
            alt.Y('count()', title=y_axis_title)
        ).properties(
            title=alt.Title(title,
                anchor="start",
                orient="bottom",
                offset=10)
        ).show()

    @staticmethod
    def plot_histogram(items: pd.DataFrame, title: str, lang: str = "en", maxbins = "auto", **kwargs):
        sns.histplot(data=items, x='value', bins=maxbins).set(title=title)

    @staticmethod
    def plot_rank(items: pd.DataFrame, title: str, lang: str = "en", plot_labels: bool = True, domain_max: int = None, scale_min: int = 10, scale_max: int = 100, sort_field: str = "rank_median", sort_order: str = "ascending", **kwargs):


        # dataframe fields used for charting
        category_fieldname = 'category_labeled' if plot_labels else 'category'
        current_domain_max = items['count'].max() + 1 if domain_max is None else domain_max

        # per-chart translation
        Charting.translator.target = lang
        rank_axis_title = 'Rank' if lang == 'en' else Charting.translator.translate('Rank')
        chart_subtitle = 'Distribution of ranks. Lower rank indicates higher importance.' if lang == 'en' else Charting.translator.translate('Distribution of ranks. Lower rank indicates higher importance.')
        total_count_label = 'Total count' if lang == 'en' else Charting.translator.translate('Total count')


        # plot mean rank and include range of rank
        current_domain_max = items['rank_max'].max() + 1 if domain_max is None else domain_max

        base = alt.Chart(data=items).encode(
            alt.X('rank_mean', title=rank_axis_title, scale=alt.Scale(domain=(0.5, current_domain_max))),
            alt.Y(f'{category_fieldname}:N', title=None, sort=alt.EncodingSortField(field=sort_field, order=sort_order)),
            # size=alt.Size('count', scale=alt.Scale(range=[scale_min, scale_max]), title=total_count_label ),
            # color=alt.Color('count', title=total_count_label).scale(scheme="viridis")  
        ).mark_square(
            color='black'
        )

       
        
        med = alt.Chart(data=items).encode(
            alt.X('rank_median', title=rank_axis_title, scale=alt.Scale(domain=(0.5, current_domain_max))),
            alt.Y(f'{category_fieldname}:N', title=None, sort=alt.EncodingSortField(field=sort_field, order=sort_order)),
        ).mark_tick(
            color='black'
        )
        med = alt.layer(med.mark_square(
            color='red'
        ))

      

        range = alt.Chart(data=items).encode(
            alt.X('rank_min', scale=alt.Scale(domain=(0.5, current_domain_max))),
            alt.X2('rank_max'),
            alt.Y(f'{category_fieldname}:N', sort=alt.EncodingSortField(field=sort_field, order=sort_order)),
        ).mark_rule(
            color='darkgreen',
        )

        color_scheme = kwargs.get('color_scale', 'lightgreyred')
        show_subtitle = kwargs.get('show_subtitle', False)

        iqr = alt.Chart(data=items).encode(
            alt.X('rank_percentile_25', scale=alt.Scale(domain=(0.5, current_domain_max))),
            alt.X2('rank_percentile_75'),
            alt.Y(f'{category_fieldname}:N', sort=alt.EncodingSortField(field=sort_field, order=sort_order)),
            color=alt.Color('count', title=total_count_label).scale(scheme=color_scheme)  
        ).mark_rect(
            stroke='black',
            size=6
        )

        top10_rank = alt.Chart(data=items).encode(
            alt.X('rank_percentile_10', scale=alt.Scale(domain=(0.5, current_domain_max))),
            alt.Y(f'{category_fieldname}:N', sort=alt.EncodingSortField(field=sort_field, order=sort_order)),            
        ).mark_tick(
            color='red'
        )

        (range + iqr + top10_rank + base + med).properties(
            title=alt.Title(title, 
                subtitle=chart_subtitle if show_subtitle else "", 
                anchor="start", 
                orient="bottom", 
                offset=10)
        ).configure_axisY(
            grid=True,
            labelLimit=0,
            titlePadding=30
        ).configure_axisX(
            grid=False,
            titlePadding=20    
        ).show()

    

    @staticmethod 
    def plot_stacked_matrix(items: pd.DataFrame, title: str, response_categories: List[str], lang: str = "en", descending_scale: bool = True, plot_labels: bool = True, domain_max: int = None, **kwargs):
        
        # color scale construction        
        category_fieldname = 'category_labeled' if plot_labels else 'category'
        response_fieldname = 'response_labeled' if plot_labels else 'response'
        domain_values = list(response_categories.values()) if plot_labels else list(response_categories.keys())

        color_scale = kwargs.get('color_scale', 'RdYlGn')
        show_subtitle = kwargs.get('show_subtitle', False)

        contains_nan = True if np.nan in response_categories.keys() else False
        color_count = len(response_categories) if not contains_nan else len(response_categories) - 1
                
        colormap = plt.cm.get_cmap(color_scale, color_count)
        if descending_scale:
            colormap = colormap.reversed()
        norm = Normalize(vmin=0, vmax=color_count)
        

        colors = [to_hex(colormap(norm(v))) for v in range(0, color_count)]
        if contains_nan:
            colors.append('#D3D3D3')  # light gray for NaN
        
        if domain_max is None:
            domain_max = items['count'].max() + 1

        # make a dictionary to specify drawing order in dataframe
        order_dict = {key: idx for idx, key in enumerate(domain_values)}
        items_copy = items.copy()
        items_copy['category_draw_order'] = items_copy[response_fieldname].map(order_dict)

        # per-chart translation
        Charting.translator.target = lang
        chart_subtitle = "Number of times an item has been ranked, irrespective of rank" if lang == 'en' else Charting.translator.translate("Number of times an item has been ranked, irrespective of rank")
        share_axis_title = 'Share of total (%)' if lang == 'en' else Charting.translator.translate('Share of total (%)')
        share_subtitle = 'Share of responses per item' if lang == 'en' else Charting.translator.translate('Share of responses per item')

        ch = alt.Chart(data=items_copy,
            title=alt.Title(title,
                subtitle=share_subtitle if show_subtitle else "",
                anchor='start',
                orient='bottom', 
                offset=20
            )
        ).encode(
            alt.X('count:Q', title=share_axis_title, axis=alt.Axis(tickMinStep=1)).stack('normalize'),
            alt.Y(f'{category_fieldname}:N', title=None, sort=alt.EncodingSortField(field=category_fieldname, order='ascending')),
            color=alt.Color(f'{response_fieldname}:N', 
                            scale=alt.Scale(domain=list(domain_values), range=colors),
                            title=''),
            order=alt.Order('category_draw_order:O', sort='ascending')
        ).mark_bar().configure_axisY(
            labelLimit=0,
        ).configure_axisX(
            titlePadding=20
        ).show()


    @staticmethod
    def plot_bars(items: pd.DataFrame, title: str, lang: str = "en", plot_labels: bool = True, domain_max: int = None, **kwargs):

        # determine category field to use
        category_fieldname = 'category_labeled' if plot_labels else 'category'

        if domain_max is None:
            domain_max = items['count'].max() + 1

        show_subtitle = kwargs.get('show_subtitle', False)
        bar_color = kwargs.get('color', 'darkgreen')

        # per-chart translation
        Charting.translator.target = lang
        count_axis_title = 'Total count' if lang == 'en' else Charting.translator.translate('Total count')
        chart_subtitle = "Number of times an item has been selected" if lang == 'en' else Charting.translator.translate("Number of times an item has been selected")

        base = alt.Chart(items,
            title=alt.Title(title,
                subtitle=chart_subtitle if show_subtitle else "",
                anchor="start",
                orient="bottom",
                offset=20)
        ).encode(   
            y=alt.Y(f'{category_fieldname}:N', sort='-x', title=None),
            x=alt.X('count:Q').title(count_axis_title).scale(domain=(0, domain_max)),
            text=alt.Text('count')
        )
        base = alt.layer(base.mark_bar(color=bar_color), base.mark_text(dx=20)).configure_axisY(
            titlePadding=30
        ).configure_axisY(
            labelLimit=0
        ).show()

    
    @staticmethod
    def plot_pie(items: pd.DataFrame, title: str, lang: str = "en", plot_labels: bool = True, **kwargs):
        
        # determine category field to use
        category_fieldname = 'category_labeled' if plot_labels else 'category'

        inner_radius = kwargs.get('inner_radius', 0)

        # per-chart translation
        Charting.translator.target = lang
        show_subtitle = kwargs.get('show_subtitle', False)
        chart_subtitle = "Number of times an item has been selected" if lang == 'en' else Charting.translator.translate("Number of times an item has been selected")

        # Create base chart with shared properties
        base = alt.Chart(items, 
            title=alt.Title(
                title, 
                subtitle=chart_subtitle if show_subtitle else "",
                anchor="middle", 
                orient="bottom",
                offset=20)
        ).encode(
            theta=alt.Theta('count', type="quantitative", stack=True),
            color=alt.Color(category_fieldname, type="nominal", title=None),            
        )
        
        # Create the pie chart arc
        arc = base.mark_arc(
            innerRadius=inner_radius,
            outerRadius=(inner_radius + 100)
        )
        
        # Create the text labels
        text = base.mark_text(
            radius=inner_radius + 115,
            fill="black"
        ).encode(
            text=alt.Text('count:Q')
        )
        
        # Layer the arc and text together
        (arc + text).resolve_scale( 
            theta="independent" 
        ).configure_legend(
            labelLimit=0
        ).show()



# this is the base class for survey items, including questions, subquestions, and answer options
class SurveyItem:
    
    def __init__(self, id: str, code: str, parent_id: str = None):
        self.id = id
        self.code = code
        self.parent_id = parent_id
        self.label = {}

    def add_label(self, lang: str, text: str):
        self.label[lang] = text

    def get_label(self, lang: str) -> str:
        return self.label.get(lang, self.code)



from .parser import LimesurvPyParser
from .survey_structure import SurveyStructure
from .survey import Survey
from .question import Question
from .answer import Answer
from .charting import Charting
from .charting import ChartType


from .common import QuestionProperties
from .item_set import ItemContainer
from .answer import Answer
from .survey_item import SurveyItem

from typing import List

class Question(SurveyItem):

    def __init__(self, qid: str, code: str, type: str, has_other_field: bool, is_mandatory: str):
        
        # initialize base class
        super().__init__(qid, code, parent_id=None)

        self.question_type = QuestionProperties.get_question_type_from_string(type)
        self.has_other_field = has_other_field
        self.is_mandatory = is_mandatory

        self.subquestion_container = ItemContainer()
        self.answer_container = ItemContainer()
        self.attributes = {}

    def is_numeric(self) -> bool:
        if self.get_attribute_value('numbers_only') is None:
            return None
        return self.get_attribute_value('numbers_only') in ['Y', 'y', '1', 1, True]

    def add_subquestion_item(self, subitem: SurveyItem):
        self.subquestion_container.add_item(subitem)

    def add_answer_item(self, answeritem: Answer):
        self.answer_container.add_item(answeritem)

    def get_answer_items(self) -> List[SurveyItem]:
        items = self.answer_container.get_all()
        return items
    
    def get_subquestion_items(self) -> List[SurveyItem]:
        items = self.subquestion_container.get_all()
        return items
    
    def get_attribute_value(self, attribute_name: str) -> str:
        if self.attributes is None:
            return None        
        return self.attributes.get(attribute_name, None)

    def __repr__(self):

        # print question identity and labels    
        str = f"Question (code='{self.code}', id='{self.id}', type='{self.get_type().value[1]}')\n"                                
        str += "  - labels:" + "\n"        
        for lang, label in self.label.items():
            str += f"    - {lang} : {label}" + "\n"
        
        # print subquestions, if any        
        str += f"  - subquestions ({len(self.subquestion_container.get_all())}):\n"                        
        for item in self.subquestion_container.get_all():
            str += f"    - code: {item.code}, id: {item.id}" + "\n"
            str += f"    - labels:" + "\n"
            for lang, label in item.label.items():
                str += f"      - {lang}: '{label}'" + "\n"

        # print answer options, if any                        
        str += f"  - answers ({len(self.answer_container.get_all())}):\n"    
        for answer in self.answer_container.get_all():
            str += f"    - answer (code: {answer.code}, id: {answer.id})" + "\n"
            str += f"      - labels:" + "\n"
            for lang, label in answer.label.items():
                str += f"        - {lang}: '{label}'" + "\n"

        str += f"  - other: {self.has_other_field}" + "\n"
        str += f"  - mandatory: {self.is_mandatory}" + "\n"

        return str + "\n" 
    
    def print_details(self, include_attributes: bool = False) -> str:
        print(self)
        if include_attributes:
            print("  - attributes:")
            for attr, value in self.attributes.items():
                print(f"    - {attr}: {value}")
            print("\n")

    def get_type(self) -> QuestionProperties.QuestionType:
        return self.question_type
    


import pandas as pd
from typing import Tuple, List
from typing import Tuple, List
import numpy as np
import deep_translator as dptr

from .question import Question
from .survey_item import SurveyItem
from .response_set import ResponseSet

from .common import QuestionProperties

class DataProcessing:

    translator = dptr.GoogleTranslator(source='auto', target='en')
    """Translator support for fields"""
    
    @staticmethod
    def get_response_columns(question: Question, responses: pd.DataFrame, lang: str = "en", drop_missing_values: bool = False) -> List[str]:
        
        items: pd.DataFrame 
        items = None

        if question.get_type() is QuestionProperties.QuestionType.RANKING:
            pass
        
        elif question.get_type() is QuestionProperties.QuestionType.MULTIPLE_CHOICE:
            pass
        
        elif question.get_type() is QuestionProperties.QuestionType.MATRIX:
            pass

        elif question.get_type() is QuestionProperties.QuestionType.LIST:
                
            # determine value counts
            items = responses[question.code].copy()

        elif question.get_type() is QuestionProperties.QuestionType.LONG_FREE_TEXT or question.get_type() is QuestionProperties.QuestionType.SHORT_FREE_TEXT:
            
            items = responses[[question.code]].copy()  
            items.rename(columns={
                question.code : 'value'
            }, inplace=True)

            # drop nan values, if requested
            if drop_missing_values:
                items = items.dropna(subset=['value'])

            # check whether we should treat texts as numbers
            if question.is_numeric():
                items['value'] = pd.to_numeric(items['value'], errors='coerce')
            else:
                # add autotranslation for texts
                items = DataProcessing.get_translations(items, source_col='value', dest_col='value_translated', lang=lang)
            
        elif question.get_type() is QuestionProperties.QuestionType.NUMERICAL_INPUT:

            items = responses[[question.code]].copy()  
           
            items.rename(columns={
                question.code : 'value'
            }, inplace=True)

            # drop nan values, if requested
            if drop_missing_values:
                items = items.dropna(subset=['value'])

            # convert to numeric
            items['value'] = pd.to_numeric(items['value'], errors='coerce')

        return items


    @staticmethod
    def get_translations(df: pd.DataFrame, source_col: str, dest_col: str, lang: str = 'en') -> pd.DataFrame:
        """Get translations

        :param df: DataFrame containing the data to translate.
        :type df: pd.DataFrame
        :param source_col: Column of the dataframe to translate.
        :type source_col: str
        :param dest_col: Column to store the translated values.
        :type dest_col: str
        :param lang: Target language for translation, defaults to 'en'
        :type lang: str, optional
        :return: DataFrame with translated values in the specified destination column.
        :rtype: pd.DataFrame
        """

        # use GTranslate for now
        DataProcessing.translator.target = lang
        translated_df = df.copy()
        translated_df[dest_col] = translated_df[source_col].apply(lambda x: DataProcessing.translator.translate(x) if pd.notna(x) else x)
        return translated_df

    @staticmethod
    def get_responses_for_text_question(question: Question, responses: pd.DataFrame, lang: str = "en", drop_missing_values: bool = True) -> ResponseSet:
        """Get responses for a text question.

        :param question: The question to get responses for.
        :type question: Question
        :param responses: The DataFrame containing all responses.
        :type responses: pd.DataFrame
        :param lang: The language to use for labels, defaults to "en"
        :type lang: str, optional
        :param drop_missing_values: Whether to drop missing values, defaults to True
        :type drop_missing_values: bool, optional
        :return: A DataFrame containing the responses for the text question.
        :rtype: pd.DataFrame
        """

        result = ResponseSet(question)
        items = DataProcessing.get_response_columns(question=question, responses=responses, lang=lang, drop_missing_values=drop_missing_values)        
        result.items = items
        
        return result

    @staticmethod
    def get_other_values(question: Question, responses: pd.DataFrame, lang: str = 'en') -> pd.DataFrame:
        """Get 'other' values for a question.

        :param question: The question to get 'other' values for.
        :type question: Question
        :param responses: The DataFrame containing all responses.
        :type responses: pd.DataFrame
        :param lang: The language to use for labels, defaults to 'en'
        :type lang: str, optional
        :return: A DataFrame containing the 'other' values.
        :rtype: pd.DataFrame
        """
        if not question.has_other_field:
            print(f"No 'other' field for {question.code}")
            return None
        else:
            print(f"Retrieve other values for {question.code}...")
            other_values = responses[[f'{question.code}[other]']].dropna()
            other_values = other_values.rename(columns={f'{question.code}[other]': 'other_value'})
            other_values = DataProcessing.get_translations(other_values, source_col='other_value', dest_col='other_value_translated', lang=lang)
            print(f"Found {len(other_values)} 'other' values for question {question.code}")
            return other_values   

    @staticmethod
    def get_responses_for_numeric_input_question(question: Question, responses: pd.DataFrame, lang: str = 'en', drop_missing_values: bool = True) -> ResponseSet:
        """Get responses for a numeric input question.

        :param question: The question to get responses for.
        :type question: Question
        :param responses: The DataFrame containing all responses.
        :type responses: pd.DataFrame
        :param lang: The language to use for labels, defaults to 'en'
        :type lang: str, optional
        :param drop_missing_values: Whether to drop missing values, defaults to True
        :type drop_missing_values: bool, optional
        :return: A DataFrame containing the responses for the numeric input question.
        :rtype: pd.DataFrame
        """

        result = ResponseSet(question)
        items = DataProcessing.get_response_columns(question=question, responses=responses, lang=lang, drop_missing_values=drop_missing_values)        
        result.items = items        
        return result

    @staticmethod
    def get_responses_for_list_question(question: Question, responses: pd.DataFrame, lang: str = 'en', drop_missing_values: bool = True) -> ResponseSet:
        """Get responses for a list question.

        :param question: The question to get responses for.
        :type question: Question
        :param responses: The DataFrame containing all responses.
        :type responses: pd.DataFrame
        :param lang: The language to use for labels, defaults to 'en'
        :type lang: str, optional
        :param drop_missing_values: Whether to drop missing values, defaults to True
        :type drop_missing_values: bool, optional
        :return: A DataFrame containing the responses for the list question.
        :rtype: pd.DataFrame
        """

        result = ResponseSet(question)
        items = DataProcessing.get_response_columns(question=question, responses=responses, lang=lang, drop_missing_values=drop_missing_values)        
        
        items = items.value_counts(dropna=drop_missing_values).reset_index().rename(columns={question.code: 'category'})

        # translation support
        DataProcessing.translator.target = lang

        # replace labels into new column
        answers = dict([(item.code, item.get_label(lang)) for item in question.get_answer_items()])
        answers[np.nan] = 'Missing values' if lang == 'en' else DataProcessing.translator.translate('Missing values')   
        answers['-oth-'] = 'Other value' if lang == 'en' else DataProcessing.translator.translate('Other value')
        items.insert(1, 'category_labeled', items['category'].map(answers).fillna(items['category']))

        result.items = items        
        result.other_items = DataProcessing.get_other_values(question=question, responses=responses, lang=lang) # treatment of other field - get 'other' responses
        
        return result


    @staticmethod
    def get_responses_for_matrix_question(question: Question, responses: pd.DataFrame, lang: str = 'en', drop_missing_values: bool = True, drop_response_categories: List[str] = None) -> ResponseSet:
        """Get responses for a matrix question.

        :param question: The question to get responses for.
        :type question: Question
        :param responses: The DataFrame containing all responses.
        :type responses: pd.DataFrame
        :param lang: The language to use for labels, defaults to 'en'
        :type lang: str, optional
        :param drop_missing_values: Whether to drop missing values, defaults to True
        :type drop_missing_values: bool, optional
        :param drop_response_categories: List of response categories to drop from results, defaults to None
        :type drop_response_categories: List[str], optional
        :return: A tuple containing the long format DataFrame and the aggregated DataFrame.
        :rtype: Tuple[ pd.DataFrame, dict]
        """

        result = ResponseSet(question)

        # translation support
        DataProcessing.translator.target = lang
        
        # matrix questions are coded one column per subquestion, with title in the form Question_code[Subquestion_code]
        # Values are the selected answer option codes
        # answers appear in the order of the XML spec file
        response_categories = dict([(item.code, item.get_label(lang)) for item in question.get_answer_items()])
        subquestions = dict([(item.code, item.get_label(lang)) for item in question.get_subquestion_items()])

        # store results 
        items = None

        # iterate over the columns, get the unique values in these columns, and the counts of unique values
        for subquestion_code in subquestions.keys():
            
            # generate the column name required to access relevant data in the pandas DataFrame
            column_name = f"{question.code}[{subquestion_code}]"
            try:
                # get value counts where category is the chosen answer, and count is the number of responses per observed answer 
                vals = responses[column_name].value_counts(dropna=drop_missing_values).reset_index().rename(columns={column_name: 'response'})
                vals.insert(0, 'category', subquestion_code)
                vals['response_count'] = vals['count'].sum()
                items = vals if items is None else pd.concat([items, vals], axis=0)                                
            
            except KeyError:
                print(f"Error fetching {column_name} for question {question.code}")

        # make a default item order
        items = items.sort_values(by=['category', 'response'], na_position='last')        

        if drop_response_categories is not None:
            items = items[~items['response'].isin(drop_response_categories)]

        # add categories for renaming
        # create new columns to replace codes with labels, but retain original codes as well 
        subquestions[np.nan] = 'Missing values' if lang == 'en' else DataProcessing.translator.translate('Missing values')
        if not drop_missing_values:
            response_categories[np.nan] = 'Missing values' if lang == 'en' else DataProcessing.translator.translate('Missing values')

        # remove categories to exclude from response_categories dict (remove the keys)
        if drop_response_categories is not None:
            for rc in drop_response_categories:
                if rc in response_categories:
                    del response_categories[rc]

        items.insert(1, 'category_labeled', items['category'].map(subquestions).fillna(items['category']))        
        items.insert(3, 'response_labeled', items['response'].map(response_categories).fillna(items['response']))

        result.items = items
        result.response_objects = response_categories

        return result

    @staticmethod
    def get_responses_for_ranking_question(question: Question, responses: pd.DataFrame, lang: str = 'en', max_ranked_items: int = None, min_item_count: int = None) -> ResponseSet:
        """Get responses for a ranking question.

        :param question: The ranking question to get responses for.
        :type question: Question
        :param responses: The DataFrame containing survey responses.
        :type responses: pd.DataFrame
        :param lang: The language to use for labels, defaults to 'en'
        :type lang: str, optional
        :return: A tuple containing the aggregated DataFrame and the long format DataFrame with raw ranks.
        :rtype: Tuple[pd.DataFrame, pd.DataFrame]
        """

        result = ResponseSet(question)

        # translation support
        DataProcessing.translator.target = lang

        # get relevant columns for the ranking question
        answer_options = dict([(item.code, item.get_label(lang)) for item in question.get_answer_items()])

        # iterate over the columns, get the unique values in these columns, and the counts of unique values
        # total_items_to_rank = len(answer_options.keys())
        total_items_to_rank = int(question.get_attribute_value('max_subquestions')) if question.get_attribute_value('max_subquestions') is not None else len(answer_options.keys())
        if max_ranked_items is not None:
            total_items_to_rank = min(total_items_to_rank, max_ranked_items)
        
        # iterate over potential ranks, consider offset by +1. Store results temporarily as list
        raw_ranks = []
        response_columns = []
        for rank in range(1, total_items_to_rank+1):

            has_column = True
            try:
                # get column of rank and get the item in the column
                column_name = f"{question.code}[{rank}]"
                vals = responses[column_name].value_counts().reset_index().rename(columns={column_name: 'category'})
                vals['rank'] = rank

                # Create long format: for each category-rank combination, create 'count' number of rows
                for _, row in vals.iterrows():
                    for _ in range(row['count']):
                        raw_ranks.append({
                            'category': row['category'],
                            'rank': row['rank']
                        })
            except:
                has_column = False # we do not have this column

            if has_column:
                response_columns.append(column_name)    

        # determine number of ranked items per respondent
        number_of_ranked_items = responses[response_columns].copy()
        number_of_ranked_items['count'] = number_of_ranked_items.notna().sum(axis=1) 
        number_of_ranked_items = number_of_ranked_items[['count']]
                
        # Convert long_df list to DataFrame
        raw_ranks = pd.DataFrame(raw_ranks)
        
        # extract relevant columns for individual responses
        rank_cols = responses[response_columns].copy()
        # rank_cols.dropna(how='all', inplace=True)
        rank_cols['rank_count'] = rank_cols.count(axis=1)

        def find_column_with_value(row, value):
            matches = row[row == value].index.tolist()
            col_name = matches[0] if matches else None
            if col_name is not None:
                return rank_cols.columns.get_loc(col_name) + 1  # Return rank (1-based index)
            else:
                return None

        for ao in answer_options.keys():
            rank_cols[ao] = rank_cols.apply(find_column_with_value, axis=1, value=ao)

        rank_cols = rank_cols[[key for key in answer_options.keys()] + ['rank_count']]

        # replace values with actual labels in requested language
        answer_options[np.nan] = 'Missing values' if lang == 'en' else DataProcessing.translator.translate('Missing values')
        if question.has_other_field:
            answer_options['other'] = 'Other value' if lang == 'en' else DataProcessing.translator.translate('Other value')
        raw_ranks.insert(1, 'category_labeled', raw_ranks['category'].map(answer_options).fillna(raw_ranks['category']))

        # define custom percentile functions
        def percentile(n):
            def percentile_(x):
                return x.quantile(n)
            percentile_.__name__ = 'percentile_{:02.0f}'.format(n*100)
            return percentile_

        items = raw_ranks.groupby('category').agg({
            'rank': ['mean', 'median', 'count', 'min', 'max', percentile(0.1), percentile(0.25), percentile(0.75), percentile(0.9)]
        })
        items.columns = ['_'.join(gp) for gp in items.columns.values]
        items.reset_index(inplace=True)
        items.rename(columns={
            'rank_count' : 'count'
        }, inplace=True)

        # filter by min item count, if requested
        if min_item_count is not None:
            items = items[items['count'] >= min_item_count]

        items['relative_count'] = items['count'] / items['count'].sum()

        # sort alphabetically by category
        items.sort_values(by="category", inplace=True)
        items.insert(1, 'category_labeled', items['category'].map(answer_options).fillna(items['category']))

        result.items = items
        result.response_objects = [raw_ranks, rank_cols, number_of_ranked_items]

        return result
    
    @staticmethod
    def get_responses_for_multiple_choice_question(question: Question, responses: pd.DataFrame, lang: str = 'en') -> ResponseSet:
        """Get responses for a multiple choice question.

        :param question: The question to get responses for.
        :type question: Question
        :param responses: The DataFrame containing all responses.
        :type responses: pd.DataFrame
        :param lang: The language to use for labels, defaults to 'en'
        :type lang: str, optional
        :return: A DataFrame containing the responses for the specified question.
        :rtype: pd.DataFrame
        """

        result = ResponseSet(question)

        # translation support
        DataProcessing.translator.target = lang

        # MC questions are coded one column per answer option, with title in the form Question_code[answer_option code]
        # Values should be Y/NaN in each column
        answers = dict([(item.code, item.get_label(lang)) for item in question.get_subquestion_items()])   
        if question.has_other_field: 
            answers['other'] = 'Other value' if lang == 'en' else DataProcessing.translator.translate('Other value')
        
        # return dict
        counts = {}

        for answer_code in answers.keys():
            try:
                column_name = f"{question.code}[{answer_code}]"
                counts[answer_code] = responses[column_name].value_counts().get('Y', 0)

            except KeyError:
                print(f"Error fetching {column_name} for question {question.code}")
            
        items = pd.DataFrame.from_dict(counts, orient='index', columns=['count']).reset_index()
        items.rename(columns={'index': 'category'}, inplace=True)        
        items.insert(1, 'category_labeled', items['category'].map(answers).fillna(items['category']))
        
        result.items = items
        result.other_items = DataProcessing.get_other_values(question=question, responses=responses, lang=lang) # treatment of other field - get 'other' responses
        
        return result


import pandas as pd
import os
from typing import Union, Tuple, List, Literal

from .common import QuestionProperties
from .parser import LimesurvPyParser
from .response_set import ResponseSet
from .survey_structure import SurveyStructure
from .data_processing import DataProcessing
from .charting import Charting, ChartType
from .question import Question

class Survey:

    def __init__(self, structure: SurveyStructure):
        # set some default values and intialize class members as needed
        self.structure = structure
        self.survey_responses = None
        
    def __repr__(self):
        str = f"Survey (structure file: '{os.path.basename(self.structure.structure_filename)}')\n"
        str += "  - titles:\n"
        for lang, title in self.structure.get_titles().items():
            str += f"    - {lang}: '{title}'\n"
        str += f"  - questions ({len(self.structure.question_set.get_all())}):\n"
        return str

    def summary(self):
        print(self)

    def get_questions(self) -> List[Question]:
        return self.structure.get_questions()

    @staticmethod
    def from_lss(structure_filename: str) -> 'Survey':
        """Import survey structure from an Limesurvey lss file that is in XML format.

        :param structure_filename: Path to the lss survey structure file.
        :type structure_filename: str
        """
        structure = LimesurvPyParser.parse(structure_filename=structure_filename)
        survey = Survey(structure=structure)
        return survey

    def list_questions(self, list_labels: bool = True):
        """List all questions in the survey."""
        print("- Questions:")
        for q in self.structure.get_questions():
            print(f"  - code: {q.code}, id: {q.id}, type: {q.get_type().value[1]}")
            if list_labels:
                for lang, label in q.label.items():
                    print(f"    - {lang} : {label}")

    def print_details(self, question_code: str = None, qid: str = None, include_attributes: bool = False):
        """Print details on a question, including subquestions and answer options. Provide either question code or question ID, code takes precedence.
        Attributes of questions will be printed at request, if available.
        
        :param question_code: Code of the question to print details for. Defaults to None.
        :type question_code: str, optional
        :param question_id: ID of the question to print details for. Defaults to None.
        :type question_id: str, optional
        """
        if question_code is None and qid is None:
            print("Please provide either question code or question ID.")
            return
        
        if question_code is None:
            question = self.structure.get_question_by_id(qid)
        else:
            question = self.structure.get_question_by_code(question_code)

        if question is None:
            print(f"Question with code '{question_code}' not found.")
            return
        
        question.print_details(include_attributes=include_attributes)

    def import_responses(self, responses_filename: str, sep: str = ','):
        """Import survey responses from a CSV file. The file should have both variables (column names) and responses (row values) coded.

        :param responses_filename: Path to the CSV file with survey responses.
        :type responses_filename: str
        :param sep: Separator used in the responses file, defaults to ','
        :type sep: str, optional
        """
        self.responses_filename = responses_filename
        print(f"Importing responses from {os.path.basename(self.responses_filename)}")
        self.raw_responses = pd.read_csv(self.responses_filename, sep=sep)
        self.survey_responses = None

    def get_responses(self) -> pd.DataFrame:
        return self.survey_responses if self.survey_responses is not None else self.raw_responses   

    @property 
    def responses(self) -> pd.DataFrame:
        return self.get_responses()
    @responses.setter
    def responses(self, df: pd.DataFrame):
        self.survey_responses = df

    def export_questions(self, questions: List[Question] = None, lang: str = "en", drop_missing_values: bool = False, target: Literal['csv', 'xlsx'] = 'xlsx', path: str = None, sep: str = ',', index: bool = False):

        # include all questions of survey, if none are provided
        if questions is None:
            questions = self.get_questions()

        # gather response sets
        response_sets = []
        for question in questions:
            rs = self.describe(question_code=question.code, lang=lang, drop_missing_values=drop_missing_values, plot=False)
            response_sets.append(rs)

        # export gathered response sets
        Survey.export(
            responses=response_sets,
            target=target,
            path=path,
            sep=sep,
            index=index
        )



    @staticmethod
    def export(responses: Union[ResponseSet, List[ResponseSet]], target: Literal['csv', 'xlsx'] = 'xlsx', path: str = None, sep: str = ',', index: bool = False):
        """Export responses of one or more questions to disk.

        :param responses: ResponseSet or list of ResponseSets to export.
        :type responses: Union[ResponseSet, List[ResponseSet]]
        :param target: File format to export to, defaults to 'xlsx'
        :type target: Literal['csv', 'xlsx'], optional
        :param path: Path of excel file to write, or to folder to write csv files into, defaults to None
        :type path: str, optional
        :param sep: Separator to use in the CSV file(s), defaults to ','
        :type sep: str, optional
        :param index: Whether to include the dataframe index in the exports, defaults to False
        :type index: bool, optional
        """

        if target == 'xlsx':
            
            if path is None:
                raise ValueError("Please provide path for Excel filename to export to.")
           
            with pd.ExcelWriter(path) as writer:
                if isinstance(responses, ResponseSet):
                    # export as single excel file
                    responses : ResponseSet
                    sheet_name = f'Sheet_{responses.question.code}'
                    df = responses.items
                    df.to_excel(writer, sheet_name=sheet_name, index=index)

                    if responses.other_items is not None:
                        other_sheet_name = f'Sheet_{responses.question.code}_other'
                        responses.other_items.to_excel(writer, sheet_name=other_sheet_name, index=True)

                else:
                    rs: ResponseSet
                    for i, rs in enumerate(responses):
                        df = rs.items            
                        sheet_name = f'Sheet_{rs.question.code}'
                        # add as sheet to existing excel file
                        df.to_excel(writer, sheet_name=sheet_name, index=index)

                        if rs.other_items is not None:
                            other_sheet_name = f'Sheet_{rs.question.code}_other'
                            rs.other_items.to_excel(writer, sheet_name=other_sheet_name, index=True)
                
                print(f"Exported responses to '{path}'")

        
        elif target == 'csv':
            
            if path is None:
                path = os.path.abspath(__file__)
                print(f"Exporting to directory: {path}")
            
            if isinstance(responses, ResponseSet):
                responses : ResponseSet
                df = responses.items
                filename = f'{responses.question.code}.csv'
                df.to_csv(os.path.join(path, filename), sep=sep, index=index)

                if responses.other_items is not None:
                    other_filename = f'{responses.question.code}_other.csv'
                    responses.other_items.to_csv(os.path.join(path, other_filename), sep=sep, index=True)
            
            else:
                rs: ResponseSet
                for i, rs in enumerate(responses):
                    df = rs.items
                    filename = f'{rs.question.code}.csv'
                    df.to_csv(os.path.join(path, filename), sep=sep, index=index)

                    if rs.other_items is not None:
                        other_filename = f'{rs.question.code}_other.csv'
                        rs.other_items.to_csv(os.path.join(path, other_filename), sep=sep, index=True)

            print(f"Exported responses to '{path}'")

        else:
            raise ValueError(f"Unsupported export target: {target}")



    def describe(self, 
                 question_code: str, 
                 responses: pd.DataFrame = None, 
                 lang='en', 
                 drop_response_codes: List[str] = None, 
                 drop_missing_values: bool = False, 
                 title = None, 
                 plot_labels: bool = True, 
                 include_question_code_in_title: bool = False, 
                 plot: bool = True, 
                 **kwargs) -> ResponseSet:
        """Extract responses for a given question, and plot responses if desired.

        :param question_code: The code of the question to describe.
        :type question_code: str
        :param responses: The DataFrame with survey responses. If None, uses the survey's responses, defaults to None.
        :type responses: pd.DataFrame, optional
        :param lang: The language to use for labels, defaults to 'en'
        :type lang: str, optional
        :param drop_missing_values: Whether to drop missing values, defaults to False
        :type drop_missing_values: bool, optional
        :param title: The title to use for the description, defaults to None
        :type title: str, optional
        :param plot_labels: Whether to plot labels or question codes, defaults to True
        :type plot_labels: bool, optional
        :param include_question_code_in_title: Whether to include the question code in the title, defaults to False
        :type include_question_code_in_title: bool, optional
        :param plot: Whether to plot the results, defaults to True
        :type plot: bool, optional
        :raises ValueError: If the question code is not found
        :raises ValueError: If the question type is not supported
        :raises ValueError: If the chart type is not supported
        :return: A ResponseSet containing the responses and associated question.
        :rtype: ResponseSet
        """

        if responses is None:
            responses = self.responses

        current_question = self.structure.get_question_by_code(question_code)   

        title = current_question.get_label(lang) if title is None else title
        if include_question_code_in_title:
            title = f"{title} ({current_question.code})"

        # select method to obtain result from type of question
        result = None
        
        if current_question.get_type() is QuestionProperties.QuestionType.RANKING:
            # consider only ranks up to a specific rank?            
            max_rank = kwargs.get('max_rank', None)
            min_count = kwargs.get('min_count', None)

            result = DataProcessing.get_responses_for_ranking_question(current_question, self.responses, lang=lang, max_ranked_items=max_rank, min_item_count=min_count)                                    

            # auto-generate sensible plots
            if plot:
                charttype_summary = kwargs.get('charttype_summary', ChartType.BAR)
                if not charttype_summary in [ChartType.BAR, ChartType.PIE]:
                    raise ValueError(f"Chart type {charttype_summary} not supported for LIST questions.")
                Charting.plot_chart(charttype_summary, current_question, result.items, title=title, lang=lang, plot_labels=plot_labels, **kwargs)
                Charting.plot_chart(ChartType.BOXPLOT, current_question, result.response_objects[2], title=title, lang=lang, plot_labels=plot_labels, **kwargs)
                Charting.plot_chart(ChartType.RANKS, current_question, result.items, title=title, lang=lang, plot_labels=plot_labels, **kwargs)
                        

        elif current_question.get_type() is QuestionProperties.QuestionType.MULTIPLE_CHOICE:
            result = DataProcessing.get_responses_for_multiple_choice_question(current_question, responses, lang=lang)
            # auto-generate sensible plots  
            if plot:
                charttype = kwargs.get('charttype', ChartType.BAR)
                if not charttype in [ChartType.BAR, ChartType.PIE]:
                    raise ValueError(f"Chart type {charttype} not supported for LIST questions.")          
                Charting.plot_chart(charttype, current_question, result.items, title=title, lang=lang, plot_labels=plot_labels, **kwargs)
            

        elif current_question.get_type() is QuestionProperties.QuestionType.MATRIX:
            result = DataProcessing.get_responses_for_matrix_question(current_question, responses, lang=lang, drop_missing_values=drop_missing_values, drop_response_categories=drop_response_codes)   
              
            # auto-generate sensible plots   
            if plot:         
                Charting.plot_chart(ChartType.STACKED_MATRIX, current_question, result.items, title=title, lang=lang, plot_labels=plot_labels, response_categories=result.response_objects, **kwargs)
            

        elif current_question.get_type() is QuestionProperties.QuestionType.LIST:
            result = DataProcessing.get_responses_for_list_question(current_question, responses, lang=lang, drop_missing_values=drop_missing_values)
            
            # auto-generate sensible plots 
            if plot:
                charttype = kwargs.get('charttype', ChartType.BAR)
                if not charttype in [ChartType.BAR, ChartType.PIE]:
                    raise ValueError(f"Chart type {charttype} not supported for LIST questions.")
                Charting.plot_chart(charttype, current_question, result.items, title=title, lang=lang, plot_labels=plot_labels, **kwargs)
            
        elif current_question.get_type() is QuestionProperties.QuestionType.NUMERICAL_INPUT:
            result = DataProcessing.get_responses_for_numeric_input_question(current_question, responses, lang=lang, drop_missing_values=drop_missing_values)

            # auto-generate sensible plots
            if plot:
                Charting.plot_chart(ChartType.HISTOGRAM, current_question, result.items, title=title, lang=lang, **kwargs)

        elif current_question.get_type() is QuestionProperties.QuestionType.LONG_FREE_TEXT or current_question.get_type() is QuestionProperties.QuestionType.SHORT_FREE_TEXT:
            result = DataProcessing.get_responses_for_text_question(current_question, responses, lang=lang, drop_missing_values=drop_missing_values)
            # auto-generate sensible plots     
            if plot:       
                if current_question.is_numeric():
                    Charting.plot_chart(ChartType.FREQUENCY, current_question, result.items, title=title, lang=lang, **kwargs)
            

        # return resultset    
        return result




















    def filter_incomplete(self, status_column: str = 'submitdate'):
        # filter responses that were started but not submitted.
        self.survey_responses = self.raw_responses[self.raw_responses[status_column].notna()]
    
    





import xml.etree.ElementTree as ET
from bs4 import BeautifulSoup

from .survey_structure import SurveyStructure
from .survey_item import SurveyItem
from .question import Question
from .item_set import ItemContainer
from .answer import Answer

class LimesurvPyParser:

    @staticmethod
    def get_text(html_text: str) -> str:
        if html_text is None:
            return None
        soup = BeautifulSoup(html_text, 'html.parser')
        return soup.get_text()


    @staticmethod
    def find_matching_elements(root: ET.Element, query_path: str, target_element: str, target_value: str):
        matching_elements = []

        for element in root.findall(query_path): 
            subelement = element.find(target_element)
            if subelement is not None and subelement.text == target_value:
                matching_elements.append(element)
        return matching_elements

    @staticmethod
    def parse(structure_filename: str) -> SurveyStructure:
        """Parse LimeSurvey structure file (lss) from XML format to a SurveyStructure object.

        :param structure_filename: Path to the lss survey structure file.
        :type structure_filename: str
        :return: Parsed SurveyStructure object.
        :rtype: SurveyStructure
        """

        # survey structure object
        survey_structure = SurveyStructure(structure_filename=structure_filename)
        print(f"Parsing {structure_filename}")

        # parse structure XML file and get document root
        xml = ET.parse(structure_filename)
        root = xml.getroot()

        # parse survey titles
        titles = root.findall('.//surveys_languagesettings//rows//row')
        for title in titles:
            lang = title.find('surveyls_language').text
            survey_title = title.find('surveyls_title').text
            survey_structure.add_title(lang, survey_title)

        # parse all questions into target dictionary
        questions = root.findall('.//questions//rows//row')
        
        # iterate over questions and obtain relevant info from subelements
        for question in questions:
            
            qid = question.find('qid').text
            code = question.find('title').text
            type = question.find('type').text
            has_other_field = question.find('other').text
            mandatory = question.find('mandatory').text

            new_question = Question(
                qid = qid,
                code = code,
                type = type,
                has_other_field = has_other_field == 'Y',
                is_mandatory = mandatory
            )

            # find questions attributes
            try:
                attributes = LimesurvPyParser.find_matching_elements(root, ".//question_attributes//rows//row", "qid", qid)
                current_attributes = { q.find('attribute').text : q.find('value').text for q in attributes }
            except Exception as e:
                current_attributes = {}
            new_question.attributes = current_attributes

            # attempt to find labels for current question
            question_labels = LimesurvPyParser.find_matching_elements(root, './/question_l10ns//rows//row', 'qid', qid)
            for label in question_labels:
                lang = label.find('language').text
                label = label.find('question').text
                new_question.add_label(lang, LimesurvPyParser.get_text(label))

            # try to find subquestions, if any, and add to question, if any
            subitems = LimesurvPyParser.find_matching_elements(root, ".//subquestions//rows//row", "parent_qid", qid)
            for subitem in subitems:
                
                subitem_qid = subitem.find('qid').text
                subitem_code = subitem.find('title').text
                
                new_subitem = SurveyItem(
                    id = subitem_qid,
                    code = subitem_code,
                    parent_id=qid
                )
                
                # attempt to find labels for subquestion
                subitem_labels = LimesurvPyParser.find_matching_elements(root, './/question_l10ns//rows//row', 'qid', subitem_qid)
                for label in subitem_labels:
                    lang = label.find('language').text
                    label = label.find('question').text
                    new_subitem.add_label(lang, LimesurvPyParser.get_text(label))

                new_question.add_subquestion_item(new_subitem)

            # try to find answer options, if any
            answer_options = LimesurvPyParser.find_matching_elements(root, ".//answers//rows//row", "qid", qid)
            for answer_option in answer_options:
                
                aid = answer_option.find('aid').text
                code = answer_option.find('code').text
                
                new_answer = Answer(
                    aid = aid,
                    code = code,
                    qid = qid
                )

                # find answer labels
                answer_labels = LimesurvPyParser.find_matching_elements(root, './/answer_l10ns//rows//row', 'aid', aid)
                for answer_label in answer_labels:
                    lang = answer_label.find('language').text
                    label = answer_label.find('answer').text
                    new_answer.add_label(lang, LimesurvPyParser.get_text(label))

                new_question.add_answer_item(new_answer)

            # add question into target list
            survey_structure.add_question(new_question)
       

        return survey_structure